/* /////////////////////////////////////////////////////////////////////////////////////////////////
ConnectAudio v0.4 by MacGruber.
Control external applications and devices via network.

Version 0.4 - 2022-01-03
	Implemented UI allowing you to use custom hostname and port to connect to.
	Improved error handling.
	Improved DNS resolve.

Version 0.3 - 2021-02-27
	Fixed issues caused by VaM's new security measures, works with VaM 1.20.77.9 now.

Version 0.2 - 2020-10-11
	(no code changes)

Version 0.1 - 2020-09-26
	Initial release.

///////////////////////////////////////////////////////////////////////////////////////////////// */

using UnityEngine;
using UnityEngine.Events;
using System;
using System.Text;
using System.Net;
using System.Net.Sockets;

namespace MacGruber
{
	public class Connect : MVRScript
	{
		private Socket myClient;
		private IAsyncResult myResult;
		private IAsyncResult myHostEntryResult;

		private string myHostnameToUse;
		private int myPortToUse;
		private bool myIPv6ToUse;

		private IPAddress myIpAddress;
		private DateTime myLastSendTimestamp = DateTime.MinValue;
		private DateTime myWaitBeforeRetryTimestamp = DateTime.MinValue;
		private bool myConnected = false;
		private bool myTryConnect = false;

		private JSONStorableString myHost;
		private JSONStorableString myPort;
		private JSONStorableBool myUseIPv6;
		private JSONStorableBool myUseAutoConnect;
		private JSONStorableString myCommand;

		private readonly TimeSpan ALIVE_CHECK_TIME = new TimeSpan(0, 0, 1);

		private readonly byte[] myBuffer = new byte[256];

		public override void Init()
		{
			Utils.OnInitUI(CreateUIElement);

			Utils.SetupInfoText(this,
				"<color=#606060><size=40><b>Connect v0.4</b></size>\nUse VaM's trigger system to send commands via network to external applications and/or devices.\n\n" +
				"Once you had a successful connection, the plugin will try to auto-connect on scene load.</color>",
				300.0f, true
			);

			myHost = new JSONStorableString("Host", "localhost");
			myPort = new JSONStorableString("Port", "39340");
			RegisterString(myHost);
			RegisterString(myPort);
			Utils.SetupTextInput(this, "Host", myHost, false);
			Utils.SetupTextInput(this, "Port", myPort, false);
			myUseIPv6 = Utils.SetupToggle(this, "Use IPv6", false, false);
			myUseAutoConnect = Utils.SetupToggle(this, "Use Auto-Connect", false, false);
			
			Utils.SetupButton(this, "Connect", ConnectAction, false);
			Utils.SetupAction(this, "Connect", ConnectAction);

			myCommand = new JSONStorableString("Send", string.Empty, Send);
			RegisterString(myCommand);
		}

		private void ConnectAction()
		{
			myWaitBeforeRetryTimestamp = DateTime.MinValue;
			OnDisable();
			myTryConnect = true;
		}

		private void Update()
		{
			try
			{
				if (myWaitBeforeRetryTimestamp > DateTime.Now)
				{
					// wait
				}
				else if (myClient == null)
				{
					if (myHostEntryResult == null)
					{
						myConnected = false;
						if (!myTryConnect && !myUseAutoConnect.val)
							return;
						myTryConnect = false;

						myHostnameToUse = myHost.val;
						if (myHostnameToUse.Length == 0)
						{
							myUseAutoConnect.val = false;
							return;
						}

						if (!int.TryParse(myPort.val, out myPortToUse) || myPortToUse < 0 || myPortToUse > 65535)
						{
							LogMessage("Invalid port '{0}'.", myPort.val);
							myUseAutoConnect.val = false;
							return;
						}

						myIPv6ToUse = myUseIPv6.val;
						LogMessage("Resolving host {0}...", myHostnameToUse);
						myHostEntryResult = Dns.BeginGetHostEntry(myHostnameToUse, null, null);
					}
					else if (myHostEntryResult.IsCompleted)
					{
						IAsyncResult result = myHostEntryResult;
						myHostEntryResult = null;
						IPHostEntry hostEntry = Dns.EndGetHostEntry(result);

						if (hostEntry != null)
						{
							AddressFamily addressFamily = myIPv6ToUse ? AddressFamily.InterNetworkV6 : AddressFamily.InterNetwork;
							myIpAddress = Array.Find(hostEntry.AddressList, ip => ip.AddressFamily == addressFamily);
						}
						if (myIpAddress == null)
						{
							LogMessage("Couldn't find host '{0}' via {1}.", myHostnameToUse, myIPv6ToUse ? "IPv6" : "IPv4");
							myWaitBeforeRetryTimestamp = DateTime.Now.AddSeconds(5);
							return;
						}
						LogMessage("Trying to connect to {0}:{1}...", myIpAddress, myPortToUse);
						myClient = new Socket(myIpAddress.AddressFamily, SocketType.Stream, ProtocolType.Tcp);
						myClient.SetSocketOption(SocketOptionLevel.Socket, SocketOptionName.KeepAlive, true);
						myResult = myClient.BeginConnect(new IPEndPoint(myIpAddress, myPortToUse), null, null);
					}
				}
				else if (myConnected)
				{
					if (!myClient.Connected)
					{
						myConnected = false;
						myClient.Close();
						myClient = null;
						LogMessage("Disconnected.");
					}
					else
					{
						if (myLastSendTimestamp + ALIVE_CHECK_TIME < DateTime.Now)
						{
							// send message of length 0 to check if the connection is still there
							myBuffer[0] = 0;
							myClient.Send(myBuffer, 1, SocketFlags.None);
							myLastSendTimestamp = DateTime.Now;
						}
					}
				}
				else if (myResult == null)
				{
					myResult = myClient.BeginConnect(new IPEndPoint(myIpAddress, myPortToUse), null, null);
				}
				else if (myResult.AsyncWaitHandle.WaitOne(0, true))
				{
					if (myClient.Connected)
					{
						LogMessage("Connected!");
						myClient.EndConnect(myResult);
						myLastSendTimestamp = DateTime.Now;
						myUseAutoConnect.val = true;
						myConnected = true;
					}
					myResult = null;
				}
			}
			catch (SocketException e)
			{
				switch (e.SocketErrorCode)
				{
					case SocketError.HostNotFound:
					{
						LogMessage("Couldn't find host '{0}'.", myHostnameToUse);
						break;
					}
					default:
					{
						string msg = e.Message.TrimEnd('\n');
						LogMessage("{0}: {1}", e.SocketErrorCode, msg);
						break;
					}
				}
				myUseAutoConnect.val = false;
			}
			catch (Exception e)
			{
				LogError("Error has occured: " + e.ToString());
				myUseAutoConnect.val = false;
			}
		}

		private void OnDisable()
		{
			try
			{
				if (myClient != null)
				{
					if (myClient.Connected)
						myClient.Shutdown(SocketShutdown.Both);
					myClient.Close();
					myClient = null;
					LogMessage("Disconnected.");
				}
			}
			catch (SocketException e)
			{
				string msg = e.Message.TrimEnd('\n');
				LogMessage("{0}: {1}", e.SocketErrorCode, msg);
			}
			catch (Exception e)
			{
				LogError("Error has occured: " + e.ToString());
			}
		}

		private void OnDestroy()
		{
			Utils.OnDestroyUI();
		}

		private void Send(string message)
		{
			try
			{
				myCommand.valNoCallback = string.Empty;
				if (myClient == null || !myConnected)
					return;
				if (message.Length > 255)
				{
					LogMessage("Message too long! (max 255 characters)");
					return;
				}
				int written = Encoding.ASCII.GetBytes(message, 0, message.Length, myBuffer, 1);
				// first byte indicates length of the message
				myBuffer[0] = (byte)written;
				myClient.Send(myBuffer, 1+written, SocketFlags.None);
				myLastSendTimestamp = DateTime.Now;
			}
			catch (SocketException e)
			{
				string msg = e.Message.TrimEnd('\n');
				LogMessage("{0}: {1}", e.SocketErrorCode, msg);
			}
			catch (Exception e)
			{
				LogMessage("Error has occured: " + e.ToString());
			}
		}

		private static void LogError(string msg)
		{
			SuperController.LogError("MacGruber.Connect: " + msg);
		}

		private static void LogMessage(string msg)
		{
			SuperController.LogMessage("MacGruber.Connect: " + msg);
		}

		private static void LogMessage(string msg, params object[] args)
		{
			SuperController.LogMessage("MacGruber.Connect: " + string.Format(msg, args));
		}
	}
}